import cgi
from types import GeneratorType


exists = object()
not_exists = object()


class Raw(str):
    pass


class BaseTag(object):
    tag_name = ''
    is_start = None

    @staticmethod
    def _to_string(thing):
        if isinstance(thing, Raw) or not isinstance(thing, str):
            return str(thing)
        return cgi.escape(thing)

    @property
    def opening_str(self):
        return '<{tag_name}{attributes}>'.format(
            tag_name=self.tag_name,
            attributes=self._attributes_str,
        )

    @property
    def closing_str(self):
        return '</{tag_name}>'.format(tag_name=self.tag_name)

    @property
    def _attributes_str(self):
        if not self.attributes:
            return ''
        return ' ' + ' '.join(
            '' if value is not_exists else
            '{}{}'.format(
                name.replace('class_', 'class').replace('_', '-'),
                '' if value is exists else '="{}"'.format(self._to_string(value)))
            for name, value in self.attributes.iteritems()
        )

    @property
    def inner_str(self):
        return ''.join(self._to_string(child) for child in self.children)

    def __init__(self, *args, **attributes):
        self.children = []
        self._add_children(args)
        self.attributes = attributes

    def __call__(self, *args, **attributes):
        self._add_children(args)
        self.attributes.update(attributes)
        return self

    def _add_children(self, args):
        if isinstance(args, (list, tuple, GeneratorType)):
            for child in args:
                self._add_children(child)
        else:
            if isinstance(args, BaseTag):
                args.parent = self
            self.children.append(args)

    def __str__(self):
        return self.opening_str + self.inner_str + self.closing_str

    def __repr__(self):
        return '<{}{}>'.format(
            '/' if self.is_start is False else '',
            self.tag_name.capitalize())

    def __eq__(self, other):
        return str(self) == str(other)


def make_tag_class(name, base_class=BaseTag):
    return type(name.capitalize(), (base_class, ), {'tag_name': name})


def make_tag_classes(names, base_class=BaseTag):
    return [make_tag_class(name, base_class) for name in names]


class T(object):
    """Wrap value so that None values and missing attrs return `default`."""
    def __init__(self, value, default=''):
        self.value = value
        self.default = default

    def __getattr__(self, item):
        try:
            return T(getattr(self.value, item), self.default)
        except AttributeError:
            return T(self.default)

    def __str__(self):
        if self.value is None:
            return str(self.default)
        return str(self.value)



from . import (
    BaseTag as _BaseTag,
    Raw as _Raw,
    make_tag_classes as _make_tag_classes
)

base_tag_class = _BaseTag


class Empty(base_tag_class):
    opening_str = ''
    closing_str = ''
    _class_name = 'Empty'


br = _Raw('<br>')

# from https://www.w3schools.com/TAGs/
tag_strs = (
    'a',
    'abbr',
    'acronym',
    'address',
    'applet',
    'area',
    'article',
    'aside',
    'audio',
    'b',
    'base',
    'basefont',
    'bdi',
    'bdo',
    'big',
    'blockquote',
    'body',
    'br',
    'button',
    'canvas',
    'caption',
    'center',
    'cite',
    'code',
    'col',
    'colgroup',
    'datalist',
    'dd',
    'del',
    'details',
    'dfn',
    'dialog',
    'dir',
    'div',
    'dl',
    'dt',
    'em',
    'embed',
    'fieldset',
    'figcaption',
    'figure',
    'font',
    'footer',
    'form',
    'frame',
    'frameset',
    'h1',
    'head',
    'header',
    'hr',
    'html',
    'i',
    'iframe',
    'img',
    'input',
    'ins',
    'kbd',
    'keygen',
    'label',
    'legend',
    'li',
    'link',
    'main',
    'map',
    'mark',
    'menu',
    'menuitem',
    'meta',
    'meter',
    'nav',
    'noframes',
    'noscript',
    'object',
    'ol',
    'optgroup',
    'option',
    'output',
    'p',
    'param',
    'picture',
    'pre',
    'progress',
    'q',
    'rp',
    'rt',
    'ruby',
    's',
    'samp',
    'script',
    'section',
    'select',
    'small',
    'source',
    'span',
    'strike',
    'strong',
    'style',
    'sub',
    'summary',
    'sup',
    'table',
    'tbody',
    'td',
    'textarea',
    'tfoot',
    'th',
    'thead',
    'time',
    'title',
    'tr',
    'track',
    'tt',
    'u',
    'ul',
    'var',
    'video',
)
tag_classes = _make_tag_classes(tag_strs, base_tag_class)
tag_str_to_class = {
    tag_str: class_ for tag_str, class_ in zip(tag_strs, tag_classes)}
(
    A,
    Abbr,
    Acronym,
    Address,
    Applet,
    Area,
    Article,
    Aside,
    Audio,
    B,
    Base,
    Basefont,
    Bdi,
    Bdo,
    Big,
    Blockquote,
    Body,
    Br,
    Button,
    Canvas,
    Caption,
    Center,
    Cite,
    Code,
    Col,
    Colgroup,
    Datalist,
    Dd,
    Del,
    Details,
    Dfn,
    Dialog,
    Dir,
    Div,
    Dl,
    Dt,
    Em,
    Embed,
    Fieldset,
    Figcaption,
    Figure,
    Font,
    Footer,
    Form,
    Frame,
    Frameset,
    H1,
    Head,
    Header,
    Hr,
    Html,
    I,
    Iframe,
    Img,
    Input,
    Ins,
    Kbd,
    Keygen,
    Label,
    Legend,
    Li,
    Link,
    Main,
    Map,
    Mark,
    Menu,
    Menuitem,
    Meta,
    Meter,
    Nav,
    Noframes,
    Noscript,
    Object,
    Ol,
    Optgroup,
    Option,
    Output,
    P,
    Param,
    Picture,
    Pre,
    Progress,
    Q,
    Rp,
    Rt,
    Ruby,
    S,
    Samp,
    Script,
    Section,
    Select,
    Small,
    Source,
    Span,
    Strike,
    Strong,
    Style,
    Sub,
    Summary,
    Sup,
    Table,
    Tbody,
    Td,
    Textarea,
    Tfoot,
    Th,
    Thead,
    Time,
    Title,
    Tr,
    Track,
    Tt,
    U,
    Ul,
    Var,
    Video,
) = tag_classes



from setuptools import setup

setup(name='brotz',
      version='0.1.0',
      description='Lightweight, chiller way to compose html in python',
      url='http://github.com/leontrolski/brotz',
      author='leontrolski',
      author_email='ojhrussell@gmail.com',
      license='GNU General Public License v3.0',
      packages=['brotz'],
      zip_safe=False)



from collections import deque
import re

from . import BaseTag
from .tags import Empty


def yield_until_nested(tag):
    if isinstance(tag, BaseTag) and not isinstance(tag, BaseNested):
        yield tag
        for child in tag.children:
            for n in yield_until_nested(child):
                yield n


def yield_only_nested(tag):
    if isinstance(tag, BaseNested):
        yield tag
    for child in tag.children:
        for n in yield_only_nested(child):
            yield n


def bracket(s, cond=True):
    return '[{}]'.format(s) if cond else ''


class BaseNested(Empty):
    is_list = False
    overwrite_property = 'name'

    def __init__(self, obj_name, *args, **kwargs):
        super(BaseNested, self).__init__(*args, **kwargs)

        self.obj_name = obj_name
        self.form_parents = deque()
        for i, child in enumerate(self.children):
            for n in yield_only_nested(child):
                if self.is_list:
                    n.form_parents.appendleft(i)
                n.form_parents.appendleft(self)

    @property
    def form_parent_strs(self):
        return [
            p.obj_name if isinstance(p, BaseNested) else str(p)
            for p in self.form_parents]

    def nested_name(self, name_attr, i):
        return 'BROTZ{}{}{}{}'.format(
            ''.join(bracket(p) for p in self.form_parent_strs),
            bracket(self.obj_name),
            bracket(i, self.is_list),
            bracket(name_attr),
        )

    @property
    def inner_str(self):
        reset_map = {}  # to reset `overwrite_property` after mutation
        for i, child in enumerate(self.children):
            for n in yield_until_nested(child):
                if self.overwrite_property in n.attributes:
                    reset_map[n] = n.attributes[self.overwrite_property]
                    n.attributes[self.overwrite_property] = self.nested_name(
                        n.attributes[self.overwrite_property], i)
        to_return = super(BaseNested, self).inner_str
        for k, v in reset_map.items():
            k.attributes[self.overwrite_property] = reset_map[k]
        return to_return


class Nested(BaseNested):
    def __repr__(self):
        return '<Nested: {}>'.format(self.obj_name)


class NestedList(BaseNested):
    is_list = True

    def __repr__(self):
        return '<NestedList: {}>'.format(self.obj_name)


class MagicList(list):
    """List that extends itself if the index to set is out of range."""
    def get(self, index):
        return None if index >= len(self) else self[index]

    def __setitem__(self, index, value):
        self += [None] * (index - len(self) + 1)
        super(MagicList, self).__setitem__(index, value)


def parse_post(post_dict):
    out = {}
    for name, value in post_dict.items():
        tmp = out
        if name.startswith('BROTZ'):
            keys = re.findall(r'\[(.+?)\]', name)
            for key, next_key in zip(keys, keys[1:] + [None]):
                if next_key is None:
                    tmp[key] = value
                    continue
                if key.isdigit():
                    key = int(key)
                if not tmp.get(key):
                    if next_key.isdigit():
                        tmp[key] = MagicList()
                    else:
                        tmp[key] = {}
                tmp = tmp[key]
    return out
